--[[--------------------------------------------------------------------------------------------------
MAIN
------------------------------------------------------------------------------------------------------
file: enhancedBGA
author:	Team LTW
start project: 25.03.2019 
------------------------------------------------------------------------------------------------------
Farming Simulator 19
V1.0.0.0 - 29.04.2019
------------------------------------------------------------------------------------------------------
copyright (c) Team LTW
----------------------------------------------------------------------------------------------------]]

enhancedBGA = {}
addModEventListener(enhancedBGA) 

function enhancedBGA:loadMap(name)
	print("-- loading enhancedBGA V1.0.0.0 -- (by Team LTW)")
	
	enhancedBGA:BGAfillTypes("liquidManure", 8.5, 85.0, 280.0, 55.0)
	enhancedBGA:BGAfillTypes("manure", 22.5, 82.5, 400.0, 60.0)
	enhancedBGA:BGAfillTypes("silage", 33.0, 95.8, 185.3, 52.2)
	enhancedBGA:BGAfillTypes("grass_windrow", 40, 89.2, 583.8, 54.1)
	enhancedBGA:BGAfillTypes("potato", 88.3, 94.2, 666.1, 50.7)
	self.lastFillType = true
	enhancedBGA:BGAfillTypes("sugarbeet", 91.6, 94.6, 685.8, 50.6)
end 

function enhancedBGA:deleteMap()
end 

function enhancedBGA:mouseEvent(posX, posY, isDown, isUp, button)
end 

function enhancedBGA:keyEvent(unicode, sym, modifier, isDown)
end 

function enhancedBGA:update(dt)		
end 

function enhancedBGA:draw()
end 

function enhancedBGA:feedingTime()
	if g_currentMission.environment.currentHour == 7 or g_currentMission.environment.currentHour == 19  then
		return true
	end 
	return false
end

function enhancedBGA:BGAfillTypes(fillTypeName, TS, oTS, gas_oTS, methaneContent)
	if fillTypeName ~= nil then
		if self.fillTypeData == nil then
			self.fillTypeData = {}
		end	
		self.fillTypeData[fillTypeName] = {}
		self.fillTypeData[fillTypeName].TS = TS
		self.fillTypeData[fillTypeName].oTS = oTS
		self.fillTypeData[fillTypeName].gas_oTS = gas_oTS
		self.fillTypeData[fillTypeName].methaneContent = methaneContent
--print(tostring(fillTypeName))
		if self.BGAFillTypes == nil then
			self.BGAFillTypes = tostring(fillTypeName.." ")
		elseif self.lastFillType == nil then
			self.BGAFillTypes = self.BGAFillTypes..tostring(fillTypeName.." ")
		elseif self.lastFillType == true then
			self.BGAFillTypes = self.BGAFillTypes..tostring(fillTypeName)		
		end
	end
end 

--overwrite "Bga.load"
function Bga:load(id, xmlFile, key, customEnvironment)
	self.nodeId = id
	
    local siloKey = key..".digestateSilo"
    local loadingNode = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, siloKey..".loadingStation#node"))
    if loadingNode ~= nil then
        local loadingStation = LoadingStation:new(self.isServer, self.isClient)
        if loadingStation:load(loadingNode, xmlFile, siloKey..".loadingStation", customEnvironment) then
            self.digestateSilo.loadingStation = loadingStation
        end
    end
	
    if self.digestateSilo.loadingStation == nil then
        g_logManager:warning("Could not load loading station for '%s.loadingStation'!", siloKey)
        return false
    end
	
    local storages = {}
    local i = 0
    while true do
        local storageKey = string.format("%s.storages.storage(%d)", siloKey, i)
        if not hasXMLProperty(xmlFile, storageKey) then
            break
        end
        local storageNode = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, storageKey.."#node"))
        if storageNode ~= nil then
            local storage = Storage:new(self.isServer, self.isClient)
            if storage:load(storageNode, xmlFile, storageKey) then
                table.insert(storages, storage)
                storage:setOwnerFarmId(self:getOwnerFarmId(), true)
            else
                g_logManager:warning("Could not load storage for '%s'!", storageKey)
            end
        else
            g_logManager:warning("Missing 'node' for storage '%s'!", storageKey)
        end
        i = i + 1
    end
    if #storages == 0 then
        g_logManager:warning("Missing digestate silo storages for bga!")
        return false
    end
    self.digestateSilo.storages = storages
	
    local bunkerKey = "placeable.bga.bunker"
    local unloadingNode = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, bunkerKey..".unloadingStation#node"))
    if unloadingNode ~= nil then
        local unloadingStation = BgaSellStation:new(self.isServer, self.isClient, self)
        if unloadingStation:load(unloadingNode, xmlFile, customEnvironment, bunkerKey..".unloadingStation") then
            self.bunker.unloadingStation = unloadingStation
            unloadingStation:addTargetStorage(self)
        end
    end
    if self.bunker.unloadingStation == nil then
        g_logManager:warning("Could not load unloading station for '%s.unloadingStation'!", bunkerKey)
        return false
    end
	
    self.bunker.slots = {}
    self.bunker.fillTypeToSlot = {}
	
    local i = 0
    local slotKey = string.format("%s.slot(%d)", bunkerKey, i)
    local slot = {}
    slot.capacity = 50000
    slot.fillTypes = {}
    slot.fillLevel = 0
    slot.display = DigitalDisplay:new()
    if not slot.display:load(self.nodeId, xmlFile, slotKey..".display") then
        slot.display = nil
    else
        slot.display:setValue(0)
    end
		
    local found = false
    local j = 0
    local fillTypeKey = string.format("%s.fillType(%d)", slotKey, j)

	local BGAFillTypes = g_fillTypeManager:getFillTypesByNames(enhancedBGA.BGAFillTypes, "Warning: '"..tostring(key).. "' has invalid fillType '%s'.")
	for _, fillTypeIndex in pairs(BGAFillTypes) do
--print("--"..tostring(fillTypeIndex))
		if self.bunker.fillTypeToSlot[fillTypeIndex] == nil then
			self.bunker.fillTypeToSlot[fillTypeIndex] = slot
			if slot.fillTypes[fillTypeIndex] == nil then
				found = true
				slot.fillTypes[fillTypeIndex] = {}
				slot.fillTypes[fillTypeIndex].fillLevel = 0
				slot.fillTypes[fillTypeIndex].pricePerLiter = 0 -- changed to "0"
			else
				g_logManager:warning("'%s' already used for '%s'!", g_fillTypeManager:getFillTypeNameByIndex(fillTypeIndex), fillTypeKey)
			end
		else
			g_logManager:warning("'%s' already used by another slot for '%s'!", g_fillTypeManager:getFillTypeNameByIndex(fillTypeIndex), slotKey)
		end
	end

    if found then
		table.insert(self.bunker.slots, slot)
    else
		g_logManager:warning("No fillTypes defined for slot '%s'!", slotKey)
    end

    self.samples = {}
    if self.isClient then
        self.samples.work = g_soundManager:loadSampleFromXML(xmlFile, "placeable.bga.sounds", "work", self.baseDirectory, self.nodeId, 0, AudioGroup.ENVIRONMENT, nil, nil)
    end
    g_currentMission.environment:addDayChangeListener(self)
	
-- completely new
	self.bhkw = {}
	self.bhkw.electricityPrice = 0.25 
	
	self.fermenter = {}
	self.fermenter.fillLevel = 0
	self.fermenter.capacity = 2512000	-- 2x(d=10m, h=4m)
	self.fermenter.convert = 900 		-- in l/s, 21600 per day
	
	self.bunker.feeding = 100 			-- in l/s, 20000l per day
	self.bunker.feedingScale = 100 		-- in %		
	return true
end 	

-- add to "Bga.readStream"
local oldReadStream = Bga.readStream
function Bga:readStream(streamId, connection)
	oldReadStream(self, streamId, connection)
	if connection:getIsServer() then
		self.fermenter.fillLevel = streamReadFloat32(streamId)
	end
end

-- add to "Bga.writeStream"
local oldWriteStream = Bga.writeStream
function Bga:writeStream(streamId, connection)
	oldWriteStream(self, streamId, connection)
	if not connection:getIsServer() then
		streamWriteFloat32(streamId, self.fermenter.fillLevel)
	end
end

--overwrite "Bga.readStreamUpdate"
function Bga:readUpdateStream(streamId, timestamp, connection)
	if connection:getIsServer() then
        local fillLevel = 0
        for _, slot in ipairs(self.bunker.slots) do
            slot.fillLevel = streamReadInt32(streamId)
            fillLevel = fillLevel + slot.fillLevel
            if slot.display ~= nil then
                slot.display:setValue(slot.fillLevel)
            end 
        end
        if self.isClient then
            if fillLevel > 0  and enhancedBGA:feedingTime() == true then
                if not g_soundManager:getIsSamplePlaying(self.samples.work) then
                    g_soundManager:playSample(self.samples.work)
                end 
            else
                if g_soundManager:getIsSamplePlaying(self.samples.work) then
                    g_soundManager:stopSample(self.samples.work)
                end 
            end 
        end
		self.fermenter.fillLevel = streamReadFloat32(streamId)	
    end 
end

--overwrite "Bga.writeUpdateStream"
function Bga:writeUpdateStream(streamId, connection, dirtyMask)
    if not connection:getIsServer() then
        for _, slot in ipairs(self.bunker.slots) do
            streamWriteInt32(streamId, slot.fillLevel)
        end
		streamWriteFloat32(streamId, self.fermenter.fillLevel)
    end
end

--overwrite "Bga.update"
function Bga:update(dt) -- called every frame
    if self.isClient then
		self:raiseActive()
		if enhancedBGA:feedingTime() == true then
			if not g_soundManager:getIsSamplePlaying(self.samples.work) then
				g_soundManager:playSample(self.samples.work)
			elseif g_soundManager:getIsSamplePlaying(self.samples.work) then
				g_soundManager:stopSample(self.samples.work)
			end 
		end 
	end 	
end 

--overwrite "Bga.updateTick"
function Bga:updateTick(dt) -- called every 30ms
    if self.isServer then
        self.bunker.updateTimer = self.bunker.updateTimer + dt
		
		local timestep = (15 * dt) - 1 	-- in ms
		local updateTime = 15 	
		if g_currentMission.missionInfo.timeScale < 60 then
			timestep = (30 * dt) - 1 	-- in ms
			updateTime = 30 
		elseif g_currentMission.missionInfo.timeScale > 120 then	-- ready for higher "timeScale"
			timestep = 0 				-- in ms
			updateTime = 1 
		end 	
			
        if self.bunker.updateTimer > timestep then
			local timeFactor = (g_currentMission.missionInfo.timeScale / (3600 * 30)) * updateTime 
			if enhancedBGA:feedingTime() == true then
				for _, slot in ipairs(self.bunker.slots) do
					slot.fillLevel = 0
					for _, data in pairs(slot.fillTypes) do
						if data.fillLevel > 0 then
							local bunkerDeltaLiters = math.min(self.bunker.feedingScale * self.bunker.feeding * timeFactor, data.fillLevel) 
							if self.fermenter.fillLevel <= (self.fermenter.capacity - bunkerDeltaLiters) then
								data.fillLevel = data.fillLevel - bunkerDeltaLiters 
								self.fermenter.fillLevel = self.fermenter.fillLevel + bunkerDeltaLiters								
								if bunkerDeltaLiters > 0 then
									self:raiseDirtyFlags(self.bgaDirtyFlag)
								end
							end 
						end 
						slot.fillLevel = slot.fillLevel + data.fillLevel
					end 
					if slot.display ~= nil then
						slot.display:setValue(slot.fillLevel)
					end 
				end 
			end
--print(self.fermenter.fillLevel)
			local fermenterDeltaLiters = math.min(self.fermenter.convert * timeFactor, self.fermenter.fillLevel) 
			local bioGas = fermenterDeltaLiters * 0.6
			local difficultyMultiplier = g_currentMission.missionInfo.sellPriceMultiplier			
			self.bunker.money = self.bunker.money + (bioGas * self.bhkw.electricityPrice * difficultyMultiplier) -- "self.bunker.money" is standard
--print("money: "..tostring(self.bunker.money))
			self.fermenter.fillLevel = self.fermenter.fillLevel - fermenterDeltaLiters 
--print("fermenter: "..tostring(self.fermenter.fillLevel)) 
			local converted = fermenterDeltaLiters - bioGas 
			if converted > 0 then
			self:addDigestate(converted)
			end 
			
			self.bunker.updateTimer = 0			
        end 
    end 
end 

-- add to "Bga.saveToXMLFile"
local oldSave = Bga.saveToXMLFile
function Bga:saveToXMLFile(xmlFile, key, usedModNames)
	setXMLFloat(xmlFile, key..".fermenter.fillLevel", self.fermenter.fillLevel)
	oldSave(self, xmlFile, key, usedModNames)
end

-- add to "Bga.loadFromXMLFile"
local oldLoad = Bga.loadFromXMLFile
function Bga:loadFromXMLFile(xmlFile, key)
	self.fermenter.fillLevel = getXMLFloat(xmlFile, key..".fermenter.fillLevel") or 0
	oldLoad(self, xmlFile, key)
end